from util import RemoteMap

class Window(object):
    @classmethod
    def initialize(self, window):
        window.vars = RemoteMap(lambda k: window.get_var(k),
                                lambda k, v: window.set_var(k, v))
        window.options = RemoteMap(lambda k: window.get_option(k),
                                   lambda k, v: window.set_option(k, v))

    @property
    def buffer(self):
        return self.get_buffer()

    @property
    def cursor(self):
        return self.get_cursor()

    @cursor.setter
    def cursor(self, pos):
        self.set_cursor(pos)

    @property
    def height(self):
        return self.get_height()

    @height.setter
    def height(self, height):
        self.set_height(height)

    @property
    def width(self):
        return self.get_width()

    @width.setter
    def width(self, width):
        self.set_width(width)

    @property
    def number(self):
        return self._handle

    @property
    def row(self):
        return self.get_position()[0]

    @property
    def col(self):
        return self.get_position()[1]

    @property
    def tabpage(self):
        return self.get_tabpage()

    @property
    def valid(self):
        return self.is_valid()



from time import sleep
from random import random
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup
from threading import Thread


@with_setup(setup=cleanup)
def test_concurrent_calls():
    def produce(i):
        str = 'call send_event(%d, "integer", %d)' % (vim.channel_id, i,)
        sleep(0.05 * random())
        vim.command(str)

    count = 50
    for i in xrange(count):
        t = Thread(target=produce, args=(i,))
        t.daemon = True
        t.start()

    integers = []
    while len(integers) < count:
        integers.append(vim.next_message().arg)

    for i in xrange(count):
        ok(i in integers)


@with_setup(setup=cleanup)
def test_custom_messages():
    def produce(i):
        if i % 2 == 0:
            str = 'call send_event(%d, "nvim-message", 0)' % (vim.channel_id)
            vim.command(str)
        else:
            vim.push_message('custom-message', None)
        sleep(0.05 * random())

    count = 50
    for i in xrange(count):
        t = Thread(target=produce, args=(i,))
        t.daemon = True
        t.start()

    nvim_messages = []
    custom_messages = []
    while len(nvim_messages) < 25 or len(custom_messages) < 25 :
        message = vim.next_message()
        if message.name == 'nvim-message':
            nvim_messages.append(message)
        elif message.name == 'custom-message':
            custom_messages.append(message)

    eq(len(nvim_messages), 25)
    eq(len(custom_messages), 25)



import msgpack, logging, os, os.path
from collections import deque
from threading import Condition, Lock
from mixins import mixins
from util import VimError, VimExit, VimTimeout
import cProfile, pstats, StringIO

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)

class Promise(object):
    def __init__(self, client, request_id, expected_type=None):
        def process_response(message):
            if message.error:
                # error
                raise VimError(message.error)
            if expected_type and hasattr(client.vim, expected_type):
                # result should be a handle, wrap in it's specialized class
                klass = getattr(client.vim, expected_type)
                rv = klass(client.vim, message.result)
                klass.initialize(rv)
                return rv
            return message.result

        def wait(timeout=None):
            interrupted = False
            while True:
                if interrupted:
                    with client.interrupt_lock:
                        pass
                client.invoke_message_cb()
                with client.stream_lock:
                    if self.message:
                        return process_response(self.message)
                    interrupted = client.queue_message(timeout)

        self.message = None
        self.wait = wait

class Message(object):
    def __init__(self, client, name=None, arg=None, result=None, error=None,
                 request_id=None, response_id=None, ):
        def reply(value, error=False):
            if error:
                resp = msgpack.packb([1, request_id, value, None])
            else:
                resp = msgpack.packb([1, request_id, None, value])
            client.stream.write(resp)
        self.name = name
        self.arg = arg
        self.result = result
        self.error = error
        self.type = 'event'
        if request_id:
            self.type = 'request'
            self.reply = reply
        elif response_id:
            self.type = 'response'
            self.response_id = response_id

class Remote(object):
    """
    Base class for all remote objects(Buffer, Window...).i
    """
    def __init__(self, vim, handle):
        """
        This is the only initializer remote objects need
        """
        self._vim = vim
        self._handle = handle

    def __eq__(self, other):
        return hasattr(other, '_handle') and self._handle == other._handle


# FIFO Mutex algorithm to ensure there's no lock starvation. Adapted from
# this SO answer: http://stackoverflow.com/a/12703543/304141
class FifoLock(object):
    def __init__(self):
        self.inner = Condition(Lock())
        self.head = self.tail = 0

    def acquire(self, steal=False):
        with self.inner:
            position = self.tail
            self.tail += 1
            while position != self.head:
                self.inner.wait()

    def release(self):
        with self.inner:
            self.head += 1
            self.inner.notify_all()

    def __enter__(self):
        self.acquire()

    def __exit__(self, type, value, traceback):
        self.release()

class Client(object):
    """
    Neovim client. It depends on a stream, an object that implements three
    methods:
        - read(): Returns any amount of data as soon as it's available
        - write(chunk): Writes data
        - interrupt(): Interrupts a blocking `read()` call from another thread.

    Both methods should be fully blocking.
    """
    def __init__(self, stream, vim_compatible=False):
        self.next_request_id = 1
        self.stream = stream
        self.unpacker = msgpack.Unpacker()
        self.pending_messages = deque()
        self.pending_requests = {}
        self.vim_compatible = vim_compatible
        self.vim = None
        self.stream_lock = FifoLock()
        self.interrupt_lock = Lock()
        self.message_cb = None
        self.loop_running = False

    def unpack_message(self, timeout=None):
        """
        Unpacks the next message from the input stream. This blocks until
        `timeout` or forever if timeout=None
        """
        while True:
            try:
                debug('waiting for message...')
                msg = self.unpacker.next()
                debug('received message: %s', msg)
                name = arg = error = result = request_id = response_id = None
                msg_type = msg[0]
                if msg_type == 0:
                    request_id = msg[1]
                    name = msg[2]
                    arg = msg[3]
                elif msg_type == 1:
                    response_id = msg[1]
                    error = msg[2]
                    result = msg[3]
                elif msg_type == 2:
                    name = msg[1]
                    arg = msg[2]
                else:
                    raise Exception('Received invalid message type')
                return Message(self, name=name, arg=arg, result=result,
                               error=error, request_id=request_id,
                               response_id=response_id)
            except StopIteration:
                debug('unpacker needs more data...')
                chunk = self.stream.read(timeout)
                if not chunk:
                    if chunk == False:
                        raise VimTimeout()
                    return
                debug('feeding data to the unpacker')
                self.unpacker.feed(chunk)

    def queue_message(self, timeout=None):
        message = self.unpack_message(timeout)
        if not message:
            # interrupted
            return True
        if message.type is 'response':
            promise = self.pending_requests.pop(message.response_id)
            promise.message = message
        else:
            self.pending_messages.append(message)
        return False

    def msgpack_rpc_request(self, method_id, args, expected_type=None):
        """
        Sends a msgpack-rpc request to Neovim and return a Promise for the
        response
        """
        with self.interrupt_lock:
            self.stream.interrupt() # interrupt ongoing reads 
            with self.stream_lock:
                request_id = self.next_request_id
                # Update request id
                self.next_request_id = request_id + 1
                # Send the request
                data = msgpack.packb([0, request_id, method_id, args])
                self.stream.write(data)
                rv = Promise(self, request_id, expected_type)
                self.pending_requests[request_id] = rv
                return rv

    def next_message(self, timeout=None):
        """
        Returns the next server message
        """
        interrupted = False
        while True:
            if interrupted:
                with self.interrupt_lock:
                    pass
            self.invoke_message_cb()
            with self.stream_lock:
                if self.pending_messages:
                    return self.pending_messages.popleft()
                interrupted = self.queue_message(timeout)

    def invoke_message_cb(self):
        cb = self.message_cb
        debug('registered callback: %s', self)
        if cb:
            try:
                # If there's a callback registered for messages, invoke it
                # immediately
                cb(self.pending_messages.popleft())
            except IndexError:
                pass

    def message_loop(self, message_cb):
        profiling = 'NEOVIM_PYTHON_PROFILE' in os.environ
        try:
            assert not self.loop_running
            info('starting message loop')
            self.message_cb = message_cb
            self.loop_running = True
            if profiling:
                info('starting profiler')
                pr = cProfile.Profile()
                pr.enable()
            while True:
                try:
                    self.next_message()
                except VimExit:
                    break
        finally:
            if profiling:
                pr.disable()
                report = os.path.abspath('.nvim-python-client.profile')
                info('stopped profiler, writing report to %s', report)
                s = StringIO.StringIO()
                ps = pstats.Stats(pr, stream=s)
                ps.strip_dirs().sort_stats('tottime').print_stats(30)
                with open(report, 'w') as f:
                    f.write(s.getvalue())
            info('exiting message loop')
            self.message_cb = None
            self.loop_running = False


    def push_message(self, name, arg):
        """
        Pushes a "virtual message" that will be returned by `next_message`.
        This method can called from other threads for integration with event
        loops, such as those provided by GUI libraries.
        """
        with self.interrupt_lock:
            self.stream.interrupt()
            self.pending_messages.append(Message(self, name, arg))

    def discover_api(self):
        """
        Discovers the remote API using the special method '0'. After this
        the client will have a `vim` attribute containing an object
        that implements an interface similar to the one found in the
        python-vim module(legacy python->vim bridge)
        """
        if self.vim:
            # Only need to do this once
            return
        channel_id, api = self.msgpack_rpc_request(0, []).wait()
        api = msgpack.unpackb(api)
        # The 'Vim' class is the main entry point of the api
        classes = {'vim': type('Vim', (), {})}
        setattr(classes['vim'], 'next_message',
                lambda s, *args, **kwargs: self.next_message(*args, **kwargs))
        setattr(classes['vim'], 'message_loop',
                lambda s, *args, **kwargs: self.message_loop(*args, **kwargs))
        setattr(classes['vim'], 'push_message',
                lambda s, *args, **kwargs: self.push_message(*args, **kwargs))
        # Build classes for manipulating the remote structures, assigning to a
        # dict using lower case names as keys, so we can easily match methods
        # in the API.
        for cls in api['classes']:
            klass = type(cls + 'Base', (Remote,), {})
            # Methods of this class will pass an integer representing the
            # remote object as first argument
            classes[cls.lower()] = klass
        # now build function wrappers
        for function in api['functions']:
            # Split the name on underscores, the first part is the class name,
            # the remaining is the function name
            class_name, method_name = function['name'].split('_', 1)
            generate_wrapper(self,
                             classes[class_name],
                             method_name,
                             function['id'],
                             function['return_type'],
                             function['parameters'])
        if self.vim_compatible:
            make_vim_compatible(classes['vim'])
        # Now apply all available mixins to the generated classes
        for name, mixin in mixins.items():
            classes[name] = type(mixin.__name__, (classes[name], mixin,), {})
        # Create the 'vim object', which is a singleton of the 'Vim' class
        self.vim = classes['vim']()
        # Initialize with some useful attributes
        classes['vim'].initialize(self.vim, classes, channel_id)
        # Add attributes for each other class
        for name, klass in classes.items():
            if name != 'vim':
                setattr(self.vim, klass.__name__, klass)

def generate_wrapper(client, klass, name, fid, return_type, parameters):
    """
    Generate an API call wrapper
    """
    # Build a name->pos map for the parameters 
    parameter_names = {}
    parameter_count = 0
    for param in parameters:
        parameter_names[param[1]] = parameter_count
        parameter_count += 1
    async_name = 'send_' + name
    # These are the actual functions
    @fname(async_name)
    def async_func(*args, **kwargs):
        if isinstance(args[0], client.vim.__class__):
            # functions of the vim object don't need 'self'
            args = args[1:]
        argv = []
        # fill with positional arguments
        for i, arg in enumerate(args):
            if hasattr(client.vim, parameters[i][0]):
                # If the type is a remote object class, we use it's remote
                # handle instead
                arg = arg._handle
            # Add to the argument vector 
            argv.append(arg)
        return client.msgpack_rpc_request(fid, argv, return_type)
    @fname(name)
    def func(*args, **kwargs):
        return async_func(*args, **kwargs).wait()

    setattr(klass, async_name, async_func)
    setattr(klass, name, func)

def fname(name):
    """
    Helper for renaming generated functions
    """
    def dec(f):
        f.__name__ = name
        return f
    return dec

def process_eval_result(obj):
    def process_dict(d):
        for k, v in d.items():
            d[k] = process_value(v)
        return d

    def process_list(l):
        for i, v in enumerate(l):
            l[i] = process_value(v)
        return l

    def process_value(v):
        if isinstance(v, (int, long, float)):
            return str(v)
        if isinstance(v, dict):
            return process_dict(v)
        if isinstance(v, list):
            return process_list(v)
        return v

    return process_value(obj)

def make_vim_compatible(vim_class):
    eval_orig = vim_class.eval
    vim_class.eval = lambda *a, **ka: process_eval_result(eval_orig(*a, **ka))



from collections import deque
from util import VimExit
from signal import SIGTERM
import sys, pyuv, logging

logger = logging.getLogger(__name__)
debug, warn = (logger.debug, logger.warn,)

class UvStream(object):
    """
    Blocking read/write stream implemented with libuv. Provides the read/write
    interface required by `Client`
    """
    def __init__(self, address=None, port=None):
        debug('initializing UvStream instance')
        self._loop = pyuv.Loop()
        self._errors = deque()
        self._data = deque()
        self._written = True
        self._connected = False
        self._timed_out = False
        self._timer = pyuv.Timer(self._loop)
        self._timeout_cb = self._on_timeout.__get__(self, UvStream)
        self._read_cb = self._on_read.__get__(self, UvStream)
        self._write_cb = self._on_write.__get__(self, UvStream)
        connect_cb = self._on_connect.__get__(self, UvStream)
        # Select the type of handle
        if port:
            debug('TCP address was provided, connecting...')
            # tcp
            self._stream = pyuv.TCP(self._loop)
            self._stream.connect((address, port), connect_cb)
        elif address:
            debug('Pipe address was provided, connecting...')
            # named pipe or unix socket
            self._stream = pyuv.Pipe(self._loop)
            self._stream.connect(address, connect_cb)
        else:
            debug('No addresses were provided, will use stdin/stdout')
            # stdin/stdout
            self._read_stream = pyuv.Pipe(self._loop) 
            self._read_stream.open(sys.stdin.fileno())
            self._write_stream = pyuv.Pipe(self._loop) 
            self._write_stream.open(sys.stdout.fileno())
            self._connected = True
            self._read_stream.start_read(self._read_cb)
        async_cb = self._on_async.__get__(self, UvStream)
        self._async = pyuv.Async(self._loop, async_cb)
        self._interrupted = False
        self._term = pyuv.Signal(self._loop)
        self._term.start(self._on_term.__get__(self, UvStream), SIGTERM)

    """
    Called when the libuv stream is connected
    """
    def _on_connect(self, stream, error):
        self._loop.stop()
        if error:
            msg = pyuv.errno.strerror(error)
            warn('error connecting to neovim: %s', msg)
            self._errors.append(VimExit(msg))
            return
        self._connected = True
        self._read_stream = self._write_stream = stream
        self._read_stream.start_read(self._read_cb)


    def _on_term(self, handle, signum):
        self._loop.stop()
        self._errors.append(VimExit('Received SIGTERM'))


    """
    Called when data is read from the libuv stream
    """
    def _on_read(self, handle, data, error):
        self._loop.stop()
        if error:
            msg = pyuv.errno.strerror(error)
            warn('error reading data: %s', msg)
            self._errors.append(VimExit(msg))
            return
        elif not data:
            warn('connection was closed by neovim')
            self._errors.append(VimExit('EOF'))
            return
        else:
            debug('successfully read %d bytes of data', len(data))
            self._data.append(data)

    """
    Called when the async handle is fired
    """
    def _on_async(self, handle):
        debug('interrupted')
        self._interrupted = True
        self._loop.stop()

    """
    Called when a timeout occurs
    """
    def _on_timeout(self, handle):
        debug('timed out')
        self._timed_out = True
        self._loop.stop()

    """
    Called when data is written to the libuv stream
    """
    def _on_write(self, handle, error):
        self._loop.stop()
        if error:
            msg = pyuv.errno.strerror(error)
            warn('error writing data: %s', msg)
            self._errors.append(VimExit(msg))
            return
        debug('successfully wrote %d bytes of data', self.last_write_size)
        self._written = True
    
    """
    Runs the event loop until a certain condition
    """
    def _run(self, condition=lambda: True, timeout=None):
        if self._errors:
            debug('pending errors collected in previous event loop iteration')
            # Pending errors, throw it now
            raise self._errors.popleft()
        if timeout == 0:
            debug('0 timeout, run a non-blocking event loop iteration')
            self._loop.run(pyuv.UV_RUN_NOWAIT)
            if not condition():
                self._timed_out = True
            return

        if timeout:
            debug('prepare timer of %d seconds', timeout)
            self._timer.start(self._timeout_cb, timeout, 0)

        try:
            while not (condition() or self._timed_out):
                if self._errors:
                    debug('caught error in event loop')
                    # Error occurred, throw it to the caller
                    raise self._errors.popleft()
                # Continue processing events
                debug('run a blocking event event loop iteration...')
                self._loop.run(pyuv.UV_RUN_ONCE)

        finally:
            if timeout and not self._timed_out:
                debug('stop timer')
                self._timer.stop()

    """
    Read some data
    """
    def read(self, timeout=None):
        if self._data:
            return self._data.popleft()
        # first ensure the stream is connected
        if not self._connected:
            self._run(lambda: self._connected)
        # wait until some data is read
        self._run(lambda: self._interrupted or self._data, timeout)
        if self._timed_out:
            self._timed_out = False
            return False
        if self._interrupted:
            self._interrupted = False
            return
        # return a chunk of data
        return self._data.popleft()

    """
    Write some data
    """
    def write(self, chunk):
        if not self._connected:
            # first ensure the stream is connected
            self._run(lambda: self._connected)
        # queue the chunk for writing
        self.last_write_size = len(chunk)
        debug('writing %d bytes of data', self.last_write_size)
        self._write_stream.write(chunk, self._write_cb)
        # unset the written flag
        self._written = False
        # wait for the flag
        self._run(lambda: self._written)

    """
    Interrupts a `read` call from another thread.
    """
    def interrupt(self):
        self._async.send()



from util import RemoteMap

class Buffer(object):
    @classmethod
    def initialize(self, buffer):
        buffer.vars = RemoteMap(lambda k: buffer.get_var(k),
                                lambda k, v: buffer.set_var(k, v))
        buffer.options = RemoteMap(lambda k: buffer.get_option(k),
                                   lambda k, v: buffer.set_option(k, v))

    def __len__(self):
        return self.get_length()

    def __getitem__(self, idx):
        if not isinstance(idx, slice):
            return self.get_line(idx)
        include_end = False
        start = idx.start
        end = idx.stop
        if start == None:
            start = 0
        if end == None:
            end = -1
            include_end = True
        return self.get_slice(start, end, True, include_end)

    def __setitem__(self, idx, lines):
        if not isinstance(idx, slice):
            if lines == None:
                return self.del_line(idx)
            else:
                return self.set_line(idx, lines)
        if lines == None:
            lines = []
        include_end = False
        start = idx.start
        end = idx.stop
        if start == None:
            start = 0
        if end == None:
            end = -1
            include_end = True
        return self.set_slice(start, end, True, include_end, lines)

    def __iter__(self):
        for i in xrange(len(self)):
            yield self.get_line(i)

    def append(self, lines, index=-1):
        if isinstance(lines, basestring):
            lines = [lines]
        self.insert(index, lines)

    def mark(self, name):
        return self.get_mark(name)

    def range(self, start, end):
        return Range(self, start, end)

    @property
    def name(self):
        return self.get_name()

    @name.setter
    def name(self, value):
        return self.set_name(value)

    @property
    def number(self):
        return self.get_number()

    @property
    def valid(self):
        return self.is_valid()


class Range(object):
    def __init__(self, buffer, start, end):
        self._buffer = buffer
        self.start = start - 1
        self.end = end

    def __len__(self):
        return self.end - self.start

    def __getitem__(self, idx):
        if not isinstance(idx, slice):
            return self._buffer[self._normalize_index(idx)]
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start == None:
            start = self.start
        if end == None:
            end = self.end
        return self._buffer[start:end]

    def __setitem__(self, idx, lines):
        if not isinstance(idx, slice):
            self._buffer[self._normalize_index(idx)] = lines
            return
        start = self._normalize_index(idx.start)
        end = self._normalize_index(idx.stop)
        if start == None:
            start = self.start
        if end == None:
            end = self.end
        self._buffer[start:end] = lines

    def __iter__(self):
        for i in xrange(self.start, self.end):
            yield self._buffer[i]

    def append(self, lines, i=None):
        i = self._normalize_index(i)
        if i == None:
            i = self.end
        self._buffer.append(lines, i)

    def _normalize_index(self, index):
        if index == None:
            return None
        if index < 0:
            index = self.end - 1
        else:
            index += self.start
            if index >= self.end:
                index = self.end - 1
        return index



# -*- coding: utf-8 -*-
import os
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup

cid = vim.channel_id

@with_setup(setup=cleanup)
def test_call_and_reply():
    cmd = 'let g:result = send_call(%d, "client-call", [1,2,3])' % cid
    promise = vim.send_command(cmd)
    call = vim.next_message()
    eq(call.name, 'client-call')
    eq(call.arg, [1, 2, 3])
    call.reply([4, 5, 6])
    promise.wait()
    eq(vim.vars['result'], [4, 5, 6])


@with_setup(setup=cleanup)
def test_call_api_before_reply():
    cmd = 'let g:result = send_call(%d, "client-call", [1,2,3])' % cid
    promise = vim.send_command(cmd)
    call = vim.next_message()
    eq(call.arg, [1, 2, 3])
    vim.command('let g:result2 = [7, 8, 9]')
    call.reply(vim.vars['result2'])
    promise.wait()
    eq(vim.vars['result'], [7, 8, 9])


@with_setup(setup=cleanup)
def test_recursion():
    vim.vars['result1'] = 0
    vim.vars['result2'] = 0
    vim.vars['result3'] = 0
    vim.vars['result4'] = 0
    cmd = 'let g:result1 = send_call(%d, "client-call", %d)' % (cid, 2,)
    promise1 = vim.send_command(cmd)
    call1 = vim.next_message()
    arg = call1.arg * 2
    cmd = 'let g:result2 = send_call(%d, "client-call", %d)' % (cid, arg,)
    promise2 = vim.send_command(cmd)
    call2 = vim.next_message()
    arg = call2.arg * 2
    cmd = 'let g:result3 = send_call(%d, "client-call", %d)' % (cid, arg,)
    promise3 = vim.send_command(cmd)
    call3 = vim.next_message()
    arg = call3.arg * 2
    cmd = 'let g:result4 = send_call(%d, "client-call", %d)' % (cid, arg,)
    promise3 = vim.send_command(cmd)
    call4 = vim.next_message()
    # start responding the calls and asserting variable values
    eq(vim.vars['result4'], 0)
    call4.reply(call4.arg)
    eq(vim.vars['result4'], 16)
    eq(vim.vars['result3'], 0)
    call3.reply(call3.arg)
    eq(vim.vars['result3'], 8)
    eq(vim.vars['result2'], 0)
    call2.reply(call2.arg)
    eq(vim.vars['result2'], 4)
    eq(vim.vars['result1'], 0)
    call1.reply(call1.arg)
    eq(vim.vars['result1'], 2)



from imp import find_module, load_module
import os, sys, inspect, logging, os.path
from traceback import format_exc
from util import VimExit

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)


class RedirectStream(object):
    def __init__(self, redirect_handler):
        self.redirect_handler = redirect_handler

    def write(self, data):
        self.redirect_handler(data)

    def writelines(self, seq):
        self.redirect_handler('\n'.join(seq))


class PluginHost(object):
    """
    Class that transforms the python interpreter into a plugin host for Neovim.
    It takes care of discovering plugins and routing events/calls sent by
    Neovim to the appropriate handlers(registered by plugins)
    """
    def __init__(self, vim, discovered_plugins=[]):
        self.vim = vim
        self.method_handlers = {}
        self.event_handlers = {}
        self.discovered_plugins = discovered_plugins
        self.installed_plugins = []
        sys.modules['vim'] = vim

    def __enter__(self):
        vim = self.vim
        info('install import hook/path')
        self.hook = path_hook(vim)
        sys.path_hooks.append(self.hook)
        vim.VIM_SPECIAL_PATH = '_vim_path_'
        sys.path.append(vim.VIM_SPECIAL_PATH)
        info('redirect sys.stdout and sys.stderr')
        self.saved_stdout = sys.stdout
        self.saved_stderr = sys.stderr
        sys.stdout = RedirectStream(lambda data: vim.out_write(data))
        sys.stderr = RedirectStream(lambda data: vim.err_write(data))
        debug('installing plugins')
        self.install_plugins()
        return self

    def __exit__(self, type, value, traceback):
        for plugin in self.installed_plugins:
            if hasattr(plugin, 'on_plugin_teardown'):
                plugin.on_plugin_teardown()
        vim = self.vim
        info('uninstall import hook/path')
        sys.path.remove(vim.VIM_SPECIAL_PATH)
        sys.path_hooks.remove(self.hook)
        info('restore sys.stdout and sys.stderr')
        sys.stdout = self.saved_stdout
        sys.stderr = self.saved_stderr

    def discover_plugins(self):
        loaded = set()
        for directory in discover_runtime_directories(self.vim):
            for name in os.listdir(directory):
                if not name.startswith('nvim_'):
                    continue
                name = os.path.splitext(name)[0]
                if name in loaded:
                    continue
                loaded.add(name)
                try:
                    discovered = find_module(name, [directory])
                except:
                    err_str = format_exc(5)
                    warn('error while searching module %s: %s', name, err_str)
                    continue
                debug('discovered %s', name)
                try:
                    file, pathname, description = discovered
                    module = load_module(name, file, pathname, description)
                    for name, value in inspect.getmembers(module, inspect.isclass):
                        if name.startswith('Nvim'):
                            self.discovered_plugins.append(value)
                    debug('loaded %s', name)
                except:
                    err_str = format_exc(5)
                    warn('error while loading module %s: %s', name, err_str)
                    continue
                finally:
                    file.close()

    def install_plugins(self):
        self.discover_plugins()
        vim = self.vim
        for plugin_class in self.discovered_plugins:
            cls_name = plugin_class.__name__
            debug('inspecting class %s', plugin_class.__name__)
            try:
                plugin = plugin_class(self.vim)
            except:
                warn('constructor for %s failed. it must accept one argument',
                     cls_name)
                continue
            methods = inspect.getmembers(plugin, inspect.ismethod)
            debug('registering event handlers for %s', plugin_class.__name__)
            for name, method in methods:
                assert method.im_self == plugin
                if not name.startswith('on_'):
                    continue
                # event handler
                event_name = name[3:]
                debug('registering %s event handler', event_name)
                if event_name not in self.event_handlers:
                    self.event_handlers[event_name] = [method]
                else:
                    self.event_handlers[event_name].append(
                        method.__get__(plugin, plugin_class))

            if hasattr(plugin, 'provides') and plugin.provides:
                for provider_method_name in plugin.provides:
                    for name, method in methods:
                        debug('checking %s', name)
                        if name == provider_method_name:
                            self.method_handlers[name] = \
                                method.__get__(plugin, plugin_class)
                            try:
                                vim.register_provider(provider_method_name)
                                debug('registered %s as a %s provider',
                                      plugin_class.__name__,
                                      provider_method_name)
                            except vim.error as e:
                                warn('error while registering provider: %s', e)
                            break
            self.installed_plugins.append(plugin)


    def on_message(self, message):
        method_handlers = self.method_handlers
        event_handlers = self.event_handlers
        if message.type == 'request':
            handler = method_handlers.get(message.name, None)
            if not handler:
                msg = 'no method handlers registered for %s' % message.name
                debug(msg)
                message.reply(msg , error=True)
                return
            try:
                debug("running method handler for '%s %s'", message.name,
                      message.arg)
                if message.arg:
                    rv = handler(message.arg)
                else:
                    rv = handler()
                debug("method handler for '%s %s' returns: %s",
                      message.name,
                      message.arg,
                      rv)
                message.reply(rv)
            except Exception as e:
                if isinstance(e, VimExit):
                    raise e
                err_str = format_exc(5)
                warn("error caught while processing call '%s %s': %s",
                     message.name,
                     message.arg,
                     err_str)
                message.reply(err_str, error=True)
        elif message.type == 'event':
            handlers = event_handlers.get(message.name, None)
            if not handlers:
                debug("no event handlers registered for %s", message.name)
                return
            debug('running event handler for %s', message.name)
            try:
                for handler in handlers:
                    handler(message.arg)
            except Exception as e:
                if isinstance(e, VimExit):
                    raise e
                err_str = format_exc(5)
                warn("error caught while processing event '%s %s': %s",
                     message.name,
                     message.arg,
                     err_str)
        else:
            assert False

    def run(self):
        self.vim.message_loop(lambda m: self.on_message(m))


# This was copied/adapted from vim-python help
def path_hook(vim):
    def _get_paths():
        return discover_runtime_directories(vim)

    def _find_module(fullname, oldtail, path):
        idx = oldtail.find('.')
        if idx > 0:
            name = oldtail[:idx]
            tail = oldtail[idx+1:]
            fmr = find_module(name, path)
            module = load_module(fullname[:-len(oldtail)] + name, *fmr)
            return _find_module(fullname, tail, module.__path__)
        else:
            fmr = find_module(fullname, path)
            return load_module(fullname, *fmr)

    class VimModuleLoader(object):
        def __init__(self, module):
            self.module = module

        def load_module(self, fullname, path=None):
            return self.module

    class VimPathFinder(object):
        @classmethod
        def find_module(cls, fullname, path=None):
            try:
                return VimModuleLoader(
                    _find_module(fullname, fullname, path or _get_paths()))
            except ImportError:
                return None

        @classmethod
        def load_module(cls, fullname, path=None):
            return _find_module(fullname, fullname, path or _get_paths())

    def hook(path):
        if path == vim.VIM_SPECIAL_PATH:
            return VimPathFinder
        else:
            raise ImportError

    return hook


def discover_runtime_directories(vim):
    is_py3 = sys.version_info >= (3, 0)
    rv = []
    for path in vim.list_runtime_paths():
        if not os.path.exists(path):
            continue
        path1 = os.path.join(path, 'pythonx')
        if is_py3:
            path2 = os.path.join(path, 'python3')
        else:
            path2 = os.path.join(path, 'python2')
        if os.path.exists(path1):
            rv.append(path1)
        if os.path.exists(path2):
            rv.append(path2)
    return rv
